/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { BST, BSTNode } from './bst';
import type { AVLTreeNested, AVLTreeNodeNested, AVLTreeOptions, BinaryTreeDeleteResult, BSTNKeyOrNode, BTNCallback, KeyOrNodeOrEntry } from '../../types';
import { BTNEntry } from '../../types';
import { IBinaryTree } from '../../interfaces';
export declare class AVLTreeNode<K = any, V = any, NODE extends AVLTreeNode<K, V, NODE> = AVLTreeNodeNested<K, V>> extends BSTNode<K, V, NODE> {
    /**
     * The constructor function initializes a new instance of a class with a key and an optional value,
     * and sets the height property to 0.
     * @param {K} key - The "key" parameter is of type K, which represents the type of the key for the
     * constructor. It is used to initialize the key property of the object being created.
     * @param {V} [value] - The "value" parameter is an optional parameter of type V. It represents the
     * value associated with the key in the constructor.
     */
    constructor(key: K, value?: V);
    protected _height: number;
    /**
     * The function returns the value of the height property.
     * @returns The height of the object.
     */
    get height(): number;
    /**
     * The above function sets the value of the height property.
     * @param {number} value - The value parameter is a number that represents the new height value to be
     * set.
     */
    set height(value: number);
}
/**
 * 1. Height-Balanced: Each node's left and right subtrees differ in height by no more than one.
 * 2. Automatic Rebalancing: AVL trees rebalance themselves automatically during insertions and deletions.
 * 3. Rotations for Balancing: Utilizes rotations (single or double) to maintain balance after updates.
 * 4. Order Preservation: Maintains the binary search tree property where left child values are less than the parent, and right child values are greater.
 * 5. Efficient Lookups: Offers O(log n) search time, where 'n' is the number of nodes, due to its balanced nature.
 * 6. Complex Insertions and Deletions: Due to rebalancing, these operations are more complex than in a regular BST.
 * 7. Path Length: The path length from the root to any leaf is longer compared to an unbalanced BST, but shorter than a linear chain of nodes.
 */
export declare class AVLTree<K = any, V = any, R = BTNEntry<K, V>, NODE extends AVLTreeNode<K, V, NODE> = AVLTreeNode<K, V, AVLTreeNodeNested<K, V>>, TREE extends AVLTree<K, V, R, NODE, TREE> = AVLTree<K, V, R, NODE, AVLTreeNested<K, V, R, NODE>>> extends BST<K, V, R, NODE, TREE> implements IBinaryTree<K, V, R, NODE, TREE> {
    /**
     * This is a constructor function for an AVLTree class that initializes the tree with keys, nodes,
     * entries, or raw elements.
     * @param keysOrNodesOrEntriesOrRawElements - The `keysOrNodesOrEntriesOrRawElements` parameter is an
     * iterable object that can contain either keys, nodes, entries, or raw elements. These elements will
     * be used to initialize the AVLTree.
     * @param [options] - The `options` parameter is an optional object that can be used to customize the
     * behavior of the AVLTree. It can include properties such as `compareFn` (a function used to compare
     * keys), `allowDuplicates` (a boolean indicating whether duplicate keys are allowed), and
     * `nodeBuilder` (
     */
    constructor(keysOrNodesOrEntriesOrRawElements?: Iterable<R | KeyOrNodeOrEntry<K, V, NODE>>, options?: AVLTreeOptions<K, V, R>);
    /**
     * The function creates a new AVL tree node with the given key and value.
     * @param {K} key - The key parameter is of type K, which represents the key of the node being
     * created.
     * @param {V} [value] - The "value" parameter is an optional parameter of type V. It represents the
     * value associated with the key in the node being created.
     * @returns The method is returning a new instance of the AVLTreeNode class, casted as the generic
     * type NODE.
     */
    createNode(key: K, value?: V): NODE;
    /**
     * The function creates a new AVL tree with the specified options and returns it.
     * @param {AVLTreeOptions} [options] - The `options` parameter is an optional object that can be
     * passed to the `createTree` function. It is used to customize the behavior of the AVL tree that is
     * being created.
     * @returns a new AVLTree object.
     */
    createTree(options?: AVLTreeOptions<K, V, R>): TREE;
    /**
     * The function checks if the input is an instance of AVLTreeNode.
     * @param {R | KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntryOrRawElement - The parameter
     * `keyOrNodeOrEntryOrRawElement` can be of type `R` or `KeyOrNodeOrEntry<K, V, NODE>`.
     * @returns a boolean value indicating whether the input parameter `keyOrNodeOrEntryOrRawElement` is
     * an instance of the `AVLTreeNode` class.
     */
    isNode(keyOrNodeOrEntryOrRawElement: R | KeyOrNodeOrEntry<K, V, NODE>): keyOrNodeOrEntryOrRawElement is NODE;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     * logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (BST) has logarithmic time complexity. constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function overrides the add method of a class and inserts a key-value pair into a data
     * structure, then balances the path.
     * @param {R | KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntryOrRawElement - The parameter
     * `keyOrNodeOrEntryOrRawElement` can accept values of type `R`, `KeyOrNodeOrEntry<K, V, NODE>`, or
     * `RawElement`.
     * @param {V} [value] - The `value` parameter is an optional value that you want to associate with
     * the key or node being added to the data structure.
     * @returns The method is returning a boolean value.
     */
    add(keyOrNodeOrEntryOrRawElement: R | KeyOrNodeOrEntry<K, V, NODE>, value?: V): boolean;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function overrides the delete method of a binary tree class and performs additional operations
     * to balance the tree after deletion.
     * @param identifier - The `identifier` parameter is the value or condition used to identify the
     * node(s) to be deleted from the binary tree. It can be of any type that is compatible with the
     * binary tree's node type.
     * @param {C} callback - The `callback` parameter is a function that will be used to determine if a
     * node should be deleted or not. It is optional and has a default value of `this._DEFAULT_CALLBACK`.
     * @returns The method is returning an array of BinaryTreeDeleteResult<NODE> objects.
     */
    delete<C extends BTNCallback<NODE>>(identifier: ReturnType<C>, callback?: C): BinaryTreeDeleteResult<NODE>[];
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_swapProperties` function swaps the key, value, and height properties between two nodes in a
     * binary search tree.
     * @param {R | BSTNKeyOrNode<K, NODE>} srcNode - The `srcNode` parameter represents either a node
     * object (`NODE`) or a key-value pair (`R`) that is being swapped with another node.
     * @param {R | BSTNKeyOrNode<K, NODE>} destNode - The `destNode` parameter is either an instance of
     * `R` or an instance of `BSTNKeyOrNode<K, NODE>`.
     * @returns The method is returning the `destNodeEnsured` object if both `srcNodeEnsured` and
     * `destNodeEnsured` are truthy. Otherwise, it returns `undefined`.
     */
    protected _swapProperties(srcNode: R | BSTNKeyOrNode<K, NODE>, destNode: R | BSTNKeyOrNode<K, NODE>): NODE | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function calculates the balance factor of a node in a binary tree.
     * @param {NODE} node - The parameter "node" is of type "NODE", which likely represents a node in a
     * binary tree data structure.
     * @returns the balance factor of a given node. The balance factor is calculated by subtracting the
     * height of the left subtree from the height of the right subtree.
     */
    protected _balanceFactor(node: NODE): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function updates the height of a node in a binary tree based on the heights of its left and
     * right children.
     * @param {NODE} node - The parameter "node" represents a node in a binary tree data structure.
     */
    protected _updateHeight(node: NODE): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_balanceLL` function performs a left-left rotation to balance a binary search tree.
     * @param {NODE} A - A is a node in a binary tree.
     */
    protected _balanceLL(A: NODE): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_balanceLR` function performs a left-right rotation to balance a binary tree.
     * @param {NODE} A - A is a node in a binary tree.
     */
    protected _balanceLR(A: NODE): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `_balanceRR` performs a right-right rotation to balance a binary tree.
     * @param {NODE} A - A is a node in a binary tree.
     */
    protected _balanceRR(A: NODE): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `_balanceRL` performs a right-left rotation to balance a binary tree.
     * @param {NODE} A - A is a node in a binary tree.
     */
    protected _balanceRL(A: NODE): void;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     * logarithmic time, where "n" is the number of nodes in the tree. The method traverses the path from the inserted node to the root. constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_balancePath` function is used to update the heights of nodes and perform rotation operations
     * to restore balance in an AVL tree after inserting a node.
     * @param {R | KeyOrNodeOrEntry<K, V, NODE>} node - The `node` parameter can be of type `R` or
     * `KeyOrNodeOrEntry<K, V, NODE>`.
     */
    protected _balancePath(node: R | KeyOrNodeOrEntry<K, V, NODE>): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function replaces an old node with a new node and sets the height of the new node to be the
     * same as the old node.
     * @param {NODE} oldNode - The `oldNode` parameter represents the node that needs to be replaced in
     * the data structure.
     * @param {NODE} newNode - The `newNode` parameter is the new node that will replace the `oldNode` in
     * the data structure.
     * @returns The method is returning the result of calling the `_replaceNode` method from the
     * superclass, with the `oldNode` and `newNode` as arguments.
     */
    protected _replaceNode(oldNode: NODE, newNode: NODE): NODE;
}
